"""Functions to generate datasets for LLM training with ballmatro hands"""
import random
from datasets import Dataset
from itertools import combinations_with_replacement
from typing import List, Tuple, Generator, Dict, Any

from ballmatro.card import Card, SUITS, RANKS, MODIFIERS
from ballmatro.optimizer import brute_force_optimize
from ballmatro.score import Score

def exhaustive_generator(max_hand_size: int, seed: int = 42) -> Generator[Tuple[List[Card], Score], None, None]:
    """Generator functions for a dataset with all possible hands of a given size
    and their optimal plays using brute force optimization.
    Args:
        max_hand_size (int): The size of the hands to generate.
    Returns:
        List[Tuple[List[Card], Score]]: A list of tuples, each containing a hand and its optimal play in the form of a Score object.
    """
    # Generate all possible cards
    cards = [Card(f"{rank}{suit}{modifier}") for suit in SUITS for rank in RANKS for modifier in [""] + MODIFIERS]
    
    # Generate all combinations of the given size
    for input in combinations_with_replacement(cards, max_hand_size):
        # Find optimal play for this input
        optimal_play = brute_force_optimize(list(input))
        yield list(input), optimal_play

def random_generator(max_hand_size: int, n: int, modifiers: List[str] = None, seed: int = 42) -> Generator[Tuple[List[Card], Score], None, None]:
    """Generator function for a dataset with random hands and their optimal plays.

    Args:
        max_hand_size (int): The maximum size of the hands to generate.
        n (int): The number of random hands to generate.
        modifiers (List[str]): A list of modifiers to apply to the cards.

    Yields:
        Tuple[List[Card], Score]: A tuple containing a hand and its optimal play in the form of a Score object.
    """
    # Set the random seed for reproducibility
    random.seed(seed)
    # Ensure modifiers are set correctly
    modifiers = _get_modifiers(modifiers)
    # Find maximum ID for list of cards of size at most max_hand_size
    ncards = len(RANKS) * len(SUITS) * len(modifiers)
    max_id = sum([ncards**i for i in range(1, max_hand_size + 1)])
    for _ in range(n):
        # Generate a random hand
        input = int2cards(random.randint(1, max_id))
        # Find optimal play for this input
        optimal_play = brute_force_optimize(input)
        yield input, optimal_play

GENERATION_ALGORITHMS = {
    "exhaustive": exhaustive_generator,
    "random": random_generator,
}

def generator_to_dict(generator: Generator[Tuple[List[Card], Score], None, None]) -> Dict[str, List[Any]]:
    """Convert a generator of tuples to a generator of dictionaries.

    Args:
        generator (Generator[Tuple[List[Card], Score]]): A generator that yields tuples of input cards and their corresponding Score.

    Returns:
        Dict[str, List[Any]]: A dictionary where each key corresponds to a field in the Score object.
    """
    # Get all the data into memory
    data = list(generator)
    # Create a dictionary with the data
    dict_data = {
        "input": [str(cards) for cards, _ in data],
        "output": [str(score.played) for _, score in data],
        "score": [score.score for _, score in data],
        "hand": [score.hand.__name__ for _, score in data],
        "chips": [score.chips for _, score in data],
        "multiplier": [score.multiplier for _, score in data],
        "remaining": [str(score.remaining) for _, score in data],
    }
    return dict_data

def to_hf_dataset(generator: Generator[Tuple[List[Card], Score], None, None]) -> Dataset:
    """Convert a dataset generator to a Hugging Face dataset format.
    
    Args:
        generator (Generator[Tuple[List[Card], Score]]): A generator that yields tuples of input cards and their corresponding Score.
    
    Returns:
        Dataset: A Hugging Face dataset containing the generated data.
    """
    # Create a Hugging Face dataset from the generator
    return Dataset.from_dict(generator_to_dict(generator))

def int2cards(i: int, modifiers: List[str] = None) -> List[Card]:
    """Map from the space of integers to the space of all possible lists of cards.

    Useful for generating random datasets from a list of random integers.

    0: empty list of cards
    [1, ncards]: all possible single cards
    [1+ncards, ncards+ncards**2]: all possible pairs of cards
    [1+ncards+ncards**2, ncards+ncards**2+ncards**3]: all possible triplets of cards
    ... and so on.

    Args:
        i (int): non-negative integer to convert.
        modifiers (List[str], optional): A list of modifiers to apply to the cards. Defaults to None, which uses the global MODIFIERS.
            An empty modifier is always included.

    Returns:
        List[Card]: A list of Card objects representing the integer.
    """
    if i < 0:
        raise ValueError("Input integer must be non-negative")
    modifiers = _get_modifiers(modifiers)
    ncards = len(RANKS) * len(SUITS) * len(modifiers)

    # Find number of cards in the list to be generated
    lenlist = 0
    while i - ncards**lenlist >= 0:
        i -= ncards**lenlist
        lenlist += 1

    # Generate the list of cards
    cards = []
    for _ in range(lenlist):
        cards.append(_int2card(i % ncards, modifiers))
        i //= ncards
    return list(reversed(cards))

def _int2card(i: int, modifiers: List[str]) -> Card:
    """Map from the space of integers to the space of all possible cards.

    Useful for generating a random card from a random integer..

    Args:
        i (int): non-negative integer to convert from 0 to total number of possible cards (minus 1).

    Returns:
        Card: A Card object representing the integer.
    """
    if i < 0:
        raise ValueError("Input integer must be non-negative")
    ncards = len(RANKS) * len(SUITS) * len(modifiers)
    if i > ncards:
        raise ValueError(f"Input integer {i} exceeds maximum value {ncards-1}")

    suit = SUITS[i % len(SUITS)]
    rank = RANKS[(i // len(SUITS)) % len(RANKS)]
    modifier = modifiers[(i // (len(SUITS) * len(RANKS))) % len(modifiers)]

    return Card(f"{rank}{suit}{modifier}")

def _get_modifiers(modifiers: List[str] = None) -> List[str]:
    """Get the list of modifiers, ensuring the empty modifier is included."""
    if modifiers is None:
        return [""] + MODIFIERS  # Include empty modifier by default
    else:
        if "" not in modifiers:
            return [""] + modifiers
        return modifiers



"""Class that represents a card, and associated functions"""
from dataclasses import dataclass
from typing import List
import re

SUITS = ["♣", "♦", "♠", "♥"]
RANKS = ["2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K", "A"]
MODIFIERS = [
    "+",  # Bonus card: +30 chips
    "x",  # Mult card: +4 multiplier
    # "*",  # Wild card: can be used as a card from any suit
]
JOKER = "🂿"

@dataclass(frozen=True)
class Card:
    """Class that represents a card"""
    txt: str  # Text representation of the card

    @property
    def suit(self) -> str:
        """Return the suit of the card, or None if the card has no suit"""
        for suit in SUITS:
            if suit in self.txt:
                return suit
        return None

    @property
    def rank(self) -> str:
        """Return the rank of the card, or None if the card has no rank"""
        match = re.match(r"([2-9]|10|J|Q|K|A)", self.txt)
        if match is not None:
            return match.group(0)
        return None
    
    @property
    def rank_numeric(self) -> int:
        """Return the numeric value of the rank of the card, or None if the card has no rank"""
        return RANKS.index(self.rank) if self.rank is not None else None

    @property
    def modifier(self) -> str:
        """Return the modifier of the card, or None if the card has no modifier"""
        for modifier in MODIFIERS:
            if modifier in self.txt[-1]:
                return modifier
        return None
    
    @property
    def is_joker(self) -> bool:
        """Return True if the card is a joker, False otherwise"""
        return self.txt[0] == JOKER
    
    @property
    def joker_rule(self) -> str:
        """Return the joker rule of the card, or None if the card is not a joker"""
        if self.is_joker:
            return self.txt[1:]
        return None

    def __repr__(self) -> str:
        """Return a string representation of the card"""
        return self.txt

def parse_card_list(txt: str) -> List[Card]:
    """Transforms a list of cards in text form into a list of Card objects.

    Example input: "[♣2, ♠3, ♥4]"
    Example output: [Card("♣2"), Card("♠3"), Card("♥4")]
    """
    # Remove opening and closing brackets
    txt = txt[1:-1]
    # Remove blanks
    txt = txt.replace(" ","")
    # No cards border case
    if len(txt) == 0:
        return []
    # Split by cards
    return [Card(cardtxt) for cardtxt in txt.split(",")]



from ballmatro.card import Card
from ballmatro.hands import StraightFlush, FourOfAKind, FullHouse, Flush, Straight, ThreeOfAKind, TwoPair, Pair, HighCard, InvalidHand, find_hand

def test_straight_flush():
    cards = [Card('10♥'), Card('J♥'), Card('Q♥'), Card('K♥'), Card('A♥')]
    assert StraightFlush.check(cards)

def test_four_of_a_kind():
    cards = [Card('10♥'), Card('10♦'), Card('10♠'), Card('10♣')]
    assert FourOfAKind.check(cards)

def test_full_house():
    cards = [Card('10♥'), Card('10♦'), Card('10♠'), Card('K♣'), Card('K♦')]
    assert FullHouse.check(cards)

def test_flush():
    cards = [Card('2♥'), Card('4♥'), Card('6♥'), Card('8♥'), Card('10♥')]
    assert Flush.check(cards)

def test_straight():
    cards = [Card('10♥'), Card('J♦'), Card('Q♠'), Card('K♣'), Card('A♦')]
    assert Straight.check(cards)

def test_three_of_a_kind():
    cards = [Card('10♥'), Card('10♦'), Card('10♠')]
    assert ThreeOfAKind.check(cards)

def test_two_pair():
    cards = [Card('10♥'), Card('10♦'), Card('K♠'), Card('K♣')]
    assert TwoPair.check(cards)

def test_pair():
    cards = [Card('10♥'), Card('10♦')]
    assert Pair.check(cards)

def test_high_card():
    cards = [Card('A♥')]
    assert HighCard.check(cards)

def test_find_hand():
    assert find_hand([Card('10♥'), Card('J♥'), Card('Q♥'), Card('K♥'), Card('A♥')]) == StraightFlush
    assert find_hand([Card('10♥'), Card('10♦'), Card('10♠'), Card('10♣')]) == FourOfAKind
    assert find_hand([Card('10♥'), Card('10♦'), Card('10♠'), Card('K♣'), Card('K♦')]) == FullHouse
    assert find_hand([Card('2♥'), Card('4♥'), Card('6♥'), Card('8♥'), Card('10♥')]) == Flush
    assert find_hand([Card('10♥'), Card('J♦'), Card('Q♠'), Card('K♣'), Card('A♦')]) == Straight
    assert find_hand([Card('10♥'), Card('10♦'), Card('10♠')]) == ThreeOfAKind
    assert find_hand([Card('10♥'), Card('10♦'), Card('K♠'), Card('K♣')]) == TwoPair
    assert find_hand([Card('10♥'), Card('10♦')]) == Pair
    assert find_hand([Card('A♥')]) == HighCard
    assert find_hand([]) == InvalidHand
    assert find_hand([Card('🂿')]) == InvalidHand




"""baLLMatro possible hands and functions to identify them"""
from collections import Counter
from dataclasses import dataclass
from typing import List

from ballmatro.card import Card

@dataclass
class PokerHand:
    """Abstract class with general methods for Poker Hands scoring"""
    ncards: int  # Number of cards that make up the hand
    chips: int  # Value in chips of the hand
    multiplier: int  # Multiplier for the chips value

    @classmethod
    def check_ncards(cls, hand: List[Card]) -> bool:
        return len(hand) == cls.ncards
    
    @classmethod
    def check(cls, hand: List[Card]) -> bool:
        """Check if the hand is of the specific type"""
        if not cls.check_ncards(hand): return False
        return cls.check_specific(hand)
    
    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a specific hand"""
        pass
    
class StraightFlush(PokerHand):
    """Straight Flush: Five cards in sequence, all of the same suit"""
    ncards = 5
    chips = 100
    multiplier = 8

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a Straight Flush"""
        return Straight.check_specific(hand) and Flush.check_specific(hand)
        
class FourOfAKind(PokerHand):
    """Four of a Kind: Four cards of the same rank"""
    ncards = 4
    chips = 60
    multiplier = 7

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for Four of a Kind"""
        return len(set([card.rank for card in hand])) == 1
    
class FullHouse(PokerHand):
    """Full House: Three cards of one rank and two cards of another rank"""
    ncards = 5
    chips = 40
    multiplier = 4

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a Full House"""
        return list(Counter([card.rank for card in hand]).values()) == [3, 2]

class Flush(PokerHand):
    """Flush: Five cards of the same suit"""
    ncards = 5
    chips = 35
    multiplier = 4

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a Flush"""
        return len(set([card.suit for card in hand])) == 1
    
class Straight(PokerHand):
    """Straight: Five cards in sequence"""
    ncards = 5
    chips = 30
    multiplier = 4

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a Straight"""
        srt = sorted([card.rank_numeric for card in hand])
        for c1, c2 in zip(srt[:-1], srt[1:]):
            if c2 - c1 != 1:
                return False
        return True
    
class ThreeOfAKind(PokerHand):
    """Three of a Kind: Three cards of the same rank"""
    ncards = 3
    chips = 30
    multiplier = 3

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for Three of a Kind"""
        return len(set([card.rank for card in hand])) == 1
    
class TwoPair(PokerHand):
    """Two Pair: Two cards of one rank and two cards of another rank"""
    ncards = 4
    chips = 20
    multiplier = 2

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for Two Pair"""
        return list(Counter([card.rank for card in hand]).values()) == [2, 2]

class Pair(PokerHand):
    """Pair: Two cards of the same rank"""
    ncards = 2
    chips = 10
    multiplier = 2

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a Pair"""
        return len(set([card.rank for card in hand])) == 1
    
class HighCard(PokerHand):
    """High Card: The highest card in the hand"""
    ncards = 1
    chips = 5
    multiplier = 1

    @classmethod
    def check_specific(cls, hand: List[Card]) -> bool:
        """Check the hand for a High Card"""
        # Just check that the card is not a joker
        return not hand[0].is_joker
    
class InvalidHand(PokerHand):
    """Special PokerHand for any kind of invalid hand"""
    ncards = 0
    chips = 0
    multiplier = 0

POKER_HANDS = [StraightFlush, FourOfAKind, FullHouse, Flush, Straight, ThreeOfAKind, TwoPair, Pair, HighCard]

def find_hand(hand: List[Card]) -> PokerHand:
    """Find which poker hand has been played. Returns the PokerHand object or None if no hand is found"""
    for poker_hand in POKER_HANDS:
        if poker_hand.check(hand):
            return poker_hand
    return InvalidHand



from ballmatro.card import Card, RANKS, SUITS, MODIFIERS
from ballmatro.generators import exhaustive_generator, random_generator, to_hf_dataset, generator_to_dict, int2cards
from ballmatro.score import Score
from ballmatro.hands import InvalidHand

def test_exhaustive_generator_size1():
    # Use a small hand size for tractable test
    results = list(exhaustive_generator(1))
    # Each result is a tuple: (hand, Score)
    assert all(isinstance(hand, list) for hand, _ in results)
    assert all(isinstance(score_info, Score) for _, score_info in results)
    # Check that the number of generated hands matches the expected count
    # There are 4 suits, 13 ranks, 2 modifiers, so 4*13*3 = 156 possible cards
    assert len(results) == 156
    # Check no repetitions in the generated hands
    assert len({tuple(hand) for hand, _ in results}) == len(results)
    # Check no invalid hands
    assert all(result.hand != InvalidHand for _, result in results)

def test_exhaustive_generator_size2():
    # Use a small hand size for tractable test
    results = list(exhaustive_generator(2))
    # Check that the number of generated hands matches the expected count
    # There are 4 suits, 13 ranks, 2 modifiers, so 4*13*3 = 156 possible cards
    # The number of combinations with replacement is (n + r - 1)
    # where n is the number of items to choose from and r is the number of items to choose
    # In this case, n = 156 and r = 2 so the number of combinations is (156 + 2 - 1) choose 2 = 156 * 157 / 2 = 12246
    assert len(results) == 12246
    # Check no repetitions in the generated hands
    assert len({tuple(hand) for hand, _ in results}) == len(results)
    # Check no invalid hands
    assert all(result.hand != InvalidHand for _, result in results)

def test_random_generator_size4():
    results = list(random_generator(max_hand_size=4, n=100))
    assert len(results) == 100
    assert all(len(result[0]) <= 4 for result in results)

def test_random_generator_size8():
    results = list(random_generator(max_hand_size=8, n=123))
    assert len(results) == 123
    assert all(len(result[0]) <= 8 for result in results)

def test_random_generator_random_seed():
    results = list(random_generator(max_hand_size=5, n=150, seed=12345))
    results2 = list(random_generator(max_hand_size=5, n=150, seed=12345))
    assert results == results2  # Ensure reproducibility with the same seed

def test_random_generator_modifiers():
    # Test with modifiers
    results = list(random_generator(max_hand_size=1, n=100, modifiers=["+"]))
    assert len(results) == 100
    # Check that the cards have the expected modifiers
    for hand, _ in results:
        for card in hand:
            assert card.modifier in [None, "+"]

def test_generator_to_dict():
    # Generate a small dataset with a small generator
    dict_generator = generator_to_dict(exhaustive_generator(1))
    # Check that the dictionary has the expected keys
    assert "input" in dict_generator
    assert "output" in dict_generator
    assert "score" in dict_generator
    assert "hand" in dict_generator
    assert "chips" in dict_generator
    assert "multiplier" in dict_generator
    assert "remaining" in dict_generator

def test_hf_dataset():
    # Generate a Hugging Face dataset with a small generator
    dataset = to_hf_dataset(exhaustive_generator(1))
    # Check that the dataset has the expected columns
    assert "input" in dataset.column_names
    assert "output" in dataset.column_names
    assert "score" in dataset.column_names
    assert "hand" in dataset.column_names
    assert "chips" in dataset.column_names
    assert "multiplier" in dataset.column_names
    assert "remaining" in dataset.column_names

def test_int2cards():
    # Test conversion of integer to cards
    ncards = len(SUITS) * len(RANKS) * (len(MODIFIERS) + 1)

    cards = int2cards(0)
    assert len(cards) == 0

    cards = int2cards(1)
    assert len(cards) == 1
    assert cards[0] == Card("2♣")

    cards = int2cards(2)
    assert len(cards) == 1
    assert cards[0] == Card("2♦")

    cards = int2cards(5)
    assert len(cards) == 1
    assert cards[0] == Card("3♣")

    cards = int2cards(52)
    assert len(cards) == 1
    assert cards[0] == Card("A♥")

    cards = int2cards(53)
    assert len(cards) == 1
    assert cards[0] == Card("2♣+")

    cards = int2cards(ncards)
    assert len(cards) == 1
    assert cards[0] == Card("A♥x")

    cards = int2cards(ncards+1)
    assert len(cards) == 2
    assert cards[0] == Card("2♣")
    assert cards[1] == Card("2♣")

    cards = int2cards(ncards+2)
    assert len(cards) == 2
    assert cards[0] == Card("2♣")
    assert cards[1] == Card("2♦")

    cards = int2cards(2*ncards)
    assert len(cards) == 2
    assert cards[0] == Card("2♣")
    assert cards[1] == Card("A♥x")

    cards = int2cards(ncards+ncards**2)
    assert len(cards) == 2
    assert cards[0] == Card("A♥x")
    assert cards[1] == Card("A♥x")

    cards = int2cards(1+ncards+ncards**2)
    assert len(cards) == 3
    assert cards[0] == Card("2♣")
    assert cards[1] == Card("2♣")
    assert cards[2] == Card("2♣")

    # Now test with modifiers

    ncards = len(SUITS) * len(RANKS) * 2

    cards = int2cards(1, modifiers=["+"])
    assert len(cards) == 1
    assert cards[0] == Card("2♣")

    cards = int2cards(ncards)
    assert len(cards) == 1
    assert cards[0] == Card("A♥+")



"""Functions to score ballmatro hands"""
from dataclasses import dataclass
from datasets import Dataset
from typing import List, Tuple, Union


from ballmatro.card import Card, parse_card_list
from ballmatro.hands import find_hand, InvalidHand


CHIPS_PER_RANK = {"2": 2, "3": 3, "4": 4, "5": 5, "6": 6, "7": 7, "8": 8, "9": 9, "10": 10, "J": 10, "Q": 10, "K": 10, "A": 11}


@dataclass
class Score:
    """Class that represents the score and details of a played hand"""
    input: List[Card]  # Cards that were available for play
    played: List[Card]  # Cards played in the hand

    def __post_init__(self):
        # Find cards that were not played
        try:
            self.remaining = self._remaining_cards(self.input, self.played)
            # Find the hand that was played
            self.hand = find_hand(self.played)
        except ValueError:
            self.remaining = None
            self.hand = InvalidHand
        # Score the played cards
        self._score_played()

    def __repr__(self):
        """Return a string representation of the score info"""
        return f"Score(input={self.input}, played={self.played}, remaining={self.remaining}, hand={self.hand.__name__}, chips={self.chips}, multiplier={self.multiplier}, score={self.score})"
    
    def _remaining_cards(self, available: List[Card], played: List[Card]) -> List[Card]:
        """Returns the remaining (not played) cards after playing a hand"""
        remaining = available.copy()
        for card in played:
            # Check if the card is available
            if card not in remaining:
                raise ValueError(f"Impossible play: card {card} not in available cards")
            # Remove the card from the remaining cards
            remaining.remove(card)
        return remaining

    def _score_played(self):
        """Given a list of played cards, find their ballmatro score

        A score of 0 is attained when the hand is not recognized or the list of played cards contains cards that are not available.
        """
        # Check if the played cards were really available
        if self.remaining is None or self.hand == InvalidHand:
            self.chips = 0
            self.multiplier = 0
            self.score = 0
            return

        # Start scoring using the chips and multiplier of the hand type
        self.chips, self.multiplier = self.hand.chips, self.hand.multiplier
        # Now iterate over the cards in the order played, and score each card individually
        for card in self.played:
            self.chips, self.multiplier = _score_card(card, self.chips, self.multiplier)

        self.score = self.chips * self.multiplier

def _score_card(card: Card, chips: int, multiplier: int) -> Tuple[int, int]:
    """Applies the scoring of a single card to the current chips and multiplier"""
    # Add the chips of the card rank to the current chips
    chips += CHIPS_PER_RANK.get(card.rank, 0)
    # Apply modifiers
    if card.modifier == "+":
        chips += 30
    elif card.modifier == "x":
        multiplier += 4
    return chips, multiplier

@dataclass
class ScoreDataset:
    """Class that represents the scores obtained over a whole Ballmatro dataset"""
    dataset: Dataset  # Dataset containing the hands and optimal plays
    plays: List[Union[str, List[Card]]]  # List of plays (hands) carried out for the dataset
    scores: List[Score] = None  # Detailed Score objects for each play
    total_score: int = 0  # Total score of the plays over the whole dataset
    normalized_score: float = 0.0  # Normalized score [0,1] of the plays over the whole dataset
    invalid_hands: int = 0  # Number of invalid hands played
    normalized_invalid_hands: float = 0.0  # Fraction of invalid hands played [0,1]

    def __post_init__(self):
        # Check inputs
        if len(self.dataset) != len(self.plays):
            raise ValueError("Dataset and plays must have the same length")
        # Format plays
        self.plays = [parse_card_list(play) if isinstance(play, str) else play for play in self.plays]
        # Score the plays
        self.scores = [Score(parse_card_list(input), played) for input, played in zip(self.dataset["input"], self.plays)]
        # Compute statistics
        self.total_score = sum(score.score for score in self.scores)
        self.normalized_score = self.total_score / sum(self.dataset["score"])
        self.invalid_hands = sum(1 for score in self.scores if score.hand == InvalidHand)
        self.normalized_invalid_hands = self.invalid_hands / len(self.scores)

    def __repr__(self):
        """Return a string representation of the score info"""
        return f"ScoreDataset(total_score={self.total_score}, normalized_score={self.normalized_score}, invalid_hands={self.invalid_hands}, normalized_invalid_hands={self.normalized_invalid_hands})"



from ballmatro.card import Card
from ballmatro.score import Score, _score_card, ScoreDataset
from datasets import Dataset

def test_score_invalid_hand():
    available = [Card(txt="2♥"), Card(txt="3♦"), Card(txt="A♠")]
    played = [Card(txt="2♥"), Card(txt="A♠")]
    assert Score(available, played).score == 0

def test_score_unavailable_card():
    available = [Card(txt="2♥"), Card(txt="3♦"), Card(txt="A♠")]
    played = [Card(txt="2♥"), Card(txt="K♠")]
    assert Score(available, played).score == 0  # Card not available

def test_score_high_card():
    available = [Card(txt="2♥"), Card(txt="3♦"), Card(txt="A♠")]
    played = [Card(txt="3♦")]
    assert Score(available, played).score == 8

def test_score_pair():
    available = [Card(txt="3♥"), Card(txt="3♦"), Card(txt="A♠")]
    played = [Card(txt="3♥"), Card(txt="3♦")]
    assert Score(available, played).score == 32

def test_score_two_pair():
    available = played = [Card(txt="3♥"), Card(txt="3♦"), Card(txt="A♠"), Card(txt="A♦")]
    assert Score(available, played).score == 96

def test_score_three_of_a_kind():
    available = [Card(txt="3♥"), Card(txt="3♦"), Card(txt="3♠"), Card(txt="A♦")]
    played = [Card(txt="3♥"), Card(txt="3♦"), Card(txt="3♠")]
    assert Score(available, played).score == 117

def test_score_straight():
    available = played = [Card(txt="2♥"), Card(txt="3♦"), Card(txt="4♠"), Card(txt="5♦"), Card(txt="6♠")]
    assert Score(available, played).score == 200

def test_score_flush():
    available = played = [Card(txt="2♥"), Card(txt="3♥"), Card(txt="5♥"), Card(txt="8♥"), Card(txt="J♥")]
    assert Score(available, played).score == 252

def test_score_full_house():
    available = played = [Card(txt="3♥"), Card(txt="3♦"), Card(txt="3♠"), Card(txt="A♦"), Card(txt="A♠")]
    assert Score(available, played).score == 284

def test_score_four_of_a_kind():
    available = played = [Card(txt="3♥"), Card(txt="3♦"), Card(txt="3♠"), Card(txt="3♣")]
    assert Score(available, played).score == 504

def test_score_straight_flush():
    available = played = [Card(txt="2♥"), Card(txt="3♥"), Card(txt="4♥"), Card(txt="5♥"), Card(txt="6♥")]
    assert Score(available, played).score == 960

def test_score_card_two_hearts():
    card = Card(txt="2♥")
    chips, multiplier = _score_card(card, 0, 1)
    assert (chips, multiplier) == (2, 1)

def test_score_card_bonus():
    card = Card(txt="A♠+")
    chips, multiplier = _score_card(card, 0, 1)
    assert (chips, multiplier) == (41, 1)

def test_score_card_mult():
    card = Card(txt="K♠x")
    chips, multiplier = _score_card(card, 0, 1)
    assert (chips, multiplier) == (10, 5)

def test_scoredataset_all_valid():
    data = {
        "input": ["[3♥,3♦]", "[2♥,3♦]"],
        "score": [32, 8],
    }
    ds = Dataset.from_dict(data)
    plays = [
        [Card("3♥"), Card("3♦")],  # valid pair
        [Card("3♦")],              # high card
    ]
    score_dataset = ScoreDataset(dataset=ds, plays=plays)
    assert score_dataset.total_score == 32 + 8
    assert score_dataset.normalized_score == 1.0
    assert score_dataset.invalid_hands == 0
    assert score_dataset.normalized_invalid_hands == 0.0

def test_scoredataset_with_invalid_hand():
    data = {
        "input": ["[3♥,3♦]"],
        "score": [32],
    }
    ds = Dataset.from_dict(data)
    plays = [
        [Card("A♠")],  # not available, should be invalid
    ]
    score_dataset = ScoreDataset(dataset=ds, plays=plays)
    assert score_dataset.total_score == 0
    assert score_dataset.normalized_score == 0.0
    assert score_dataset.invalid_hands == 1
    assert score_dataset.normalized_invalid_hands == 1.0

def test_scoredataset_mixed_valid_invalid():
    data = {
        "input": ["[3♥,3♦]", "[2♥,3♦]"],
        "score": [32, 8],
    }
    ds = Dataset.from_dict(data)
    plays = [
        [Card("3♥"), Card("3♦")],  # valid
        [Card("A♠")],              # invalid
    ]
    score_dataset = ScoreDataset(dataset=ds, plays=plays)
    assert score_dataset.total_score == 32
    assert score_dataset.invalid_hands == 1
    assert score_dataset.normalized_invalid_hands == 0.5
    assert score_dataset.normalized_score == 32/40

def test_scoredataset_strings():
    """Test ScoreDataset with string inputs for plays."""
    data = {
        "input": ["[3♥,3♦]", "[2♥,3♦]"],
        "score": [32, 8],
    }
    ds = Dataset.from_dict(data)
    plays = [
        "[3♥,3♦]",  # valid pair
        "[3♦]"      # high card
    ]
    score_dataset = ScoreDataset(dataset=ds, plays=plays)
    assert score_dataset.total_score == 32 + 8
    assert score_dataset.normalized_score == 1.0
    assert score_dataset.invalid_hands == 0
    assert score_dataset.normalized_invalid_hands == 0.0



"""Functions to find the best hand in a given set of cards"""
from itertools import combinations
import math
from typing import List

from ballmatro.card import Card
from ballmatro.score import Score

def brute_force_optimize(cards: List[Card]) -> Score:
    """Find the best hand in a given set of cards using brute force"""
    best_score = -math.inf
    for i in range(1, len(cards) + 1):
        for hand in combinations(cards, i):
            score_info = Score(cards, list(hand))
            if score_info.score > best_score:
                best_score = score_info.score
                best_result = score_info
    return best_result
